@part(system, root="cmtman.mss")
@chapter(System Details)
For simple applications, you do not have to understand much about
the structure of the CMU MIDI Toolkit in order to get useful work
done.  The preceding chapters have described the existing
 application programs
as well as Moxc, a system that helps you write real-time music
programs of your own.  Details of the available calls to send and
receive MIDI are given in Appendix @ref(mpu-app).

However, you may wish to write more ambitious programs or modify
existing programs.  To do this effectively, you must understand
how the CMT programs are put together.
This chapter describes the various components
of the CMU MIDI Toolkit from a system programmer's perspective.

@section(Introduction)
In general, CMT programs are assembled from a collection of modules,
where a module is a piece of software that performs a set of
related functions.  For example, there is a module for parsing
command lines and another module for recording MIDI data.  Modules
usually consist of a single ``.c'' file and a related ``.h'' file.
In the following sections, each module is described, and references
to example uses of the modules are given.

@section(Basic MIDI interface)
All programs in CMT that send or receive MIDI use the module @i(mpu.c),
which provides an interface between C programs and the MPU-401
MIDI Processing Unit.  In theory, @i(mpu.c) could be replaced by another
software module designed for another hardware MIDI interface.

@subsection(Interface design issues)
A few words about the overall design of this interface are in order.
To begin with, @i(mpu.c) is neither a complete interface to the MPU-401
nor to MIDI.  Instead, @i(mpu.c) is an attempt at providing the
intended community of users with a rational interface that supports
experimental, real-time computer music functions.  One of the reasons
CMT comes with source code is so that if you disagree with these design
decisions, you are free to modify or extend the system to meet your
requirements.

The main areas in which @i(mpu.c) deviates from the ``conventional'' are the
absence of ``tracks'', the way in which time is handled, pitch
specification, and the lack of external synchronization.
Tracks@Index(tracks) are a concept implemented in the MPU-401 whereby
several sequences of MIDI data can be merged in real-time.  In some
commercial sequencers, the use of tracks allows the sequencer to avoid a lot
of bookkeeping and sorting when several sequences are to be played at the
same time.  In CMT, the Adagio compiler sorts its data, so tracks are not
needed to play multiple sequences together.  For example, to play two Adagio
scores simultaneously, one can normally just concatenate the files together
and run Adagio on the new file.  This approach has the advantages that an
arbitrary number of sequences can be merged, and the resulting system is
more portable (not all MIDI interfaces implement the notion of tracks).

Timing in CMT is probably the most radical departure from MIDI.  Whereas
MIDI sequencers  normally tend to talk about time in terms of beats,
CMT measures time in units of 0.01 seconds.  This is roughly the smallest
rhythmic time deviation we can perceive.  The rationale behind this decision
is that not all music is measured in beats, and some music has several
tempi going simultaneously.  If everything is converted to time in seconds,
then one can freely combine scores with different tempi and meters as well
as scores with timing notated directly in seconds.  Another timing issue
is that the MPU-401 was designed to allow the host computer to send data
in advance of its actual use.  MIDI commands can be sent to the MPU-401
with a ``time tag'' that tells the MPU-401 when to send the data to
MIDI OUT.  This is a nice feature in that it can help a slow computer
achieve tighter real-time performance.  On the other hand, it is not very
suitable for interactive real-time programs in which one normally wants
output to occur immediately after data is sent to the MPU-401.  Time
tags also have the problem that it is hard to stop a performance
immediately, because several seconds of data buffered in the MPU-401
will continue to play after the host computer stops sending data.
CMT does not use time tags; MIDI commands are send immediately.

Pitch in CMT is based on earlier computer music systems in which
middle C is represented by the number 48.  Therefore, CMT pitch
numbers are 12 less than the corresponding MIDI pitch numbers.
CMT also allows users to redefine the interpretation of pitch numbers.
Within @i(mpu.c), there is a table with two entries for each pitch number.
One of these entries specifies a pitch and the other specifies pitch bend.
When the @i(midi@ux(  )note) routine is called, it uses the table to translate
the pitch parameter into a MIDI pitch number and a pitch bend.  This
translation is normally only enabled if the ``-tune'' option followed
by a file name is specified in the command line.

Finally, CMT at present has no means for external synchronization and
cannot now be used with other sequencers or drum machines to achieve
a synchronized performance.  This is partly a consequence of the fact that
CMT does not measure time in beats, while sequencers synchronize by sending
MIDI messages to mark beats and their subdivisions.

@subsection(Interface implementation)
The @i(mpu.c) module is primarily responsible for MIDI output.  For
example, the @i(midi@ux(  )note) routine sends commands and data to the MPU-401 
to cause it to send a MIDI note-on command.  In contrast, @i(mpu.c) never
reads data directly from the MPU-401.  Input is handled by the interrupt
handler in @i(aintr.asm).

The interrupt handler@Index(interrupt handler) (@i(aintr.asm)) is invoked
whenever the MPU-401 has data ready.  The data is read and decoded to
determine what kind of data it is.  If the data is a MIDI message (e.g. MIDI
Note-on), the message is inserted into a circular buffer, which can be read
by either the @i(getbuf@Index(getbuf)) or @i(getkey@Index(getkey)) routines
in @i(mpu.c).  The buffer provides a communication path between the
interrupt handler and the non-interrupt level of CMT programs; it also helps
prevent lost messages when a burst of MIDI events arrives.  The
@i(aintr.asm) module also receives acknowledgments for commands sent by
@i(mpu.c).  Since @i(mpu.c) cannot read these acknowledgments directly (it
might read a MIDI message),
@i(mpu.c) waits for a global variable to be changed by @i(aintr.asm)
when it wants to get an acknowledgment.

A final function of @i(aintr.asm) is maintaining the time@Index(time).  In
the current implementation, time is maintained in the MPU-401 by chosing a
tempo that results in 400 ticks per second.  Whenever the 8-bit timer
overflows in the MPU-401, a timer overflow message is sent to @i(aintr.asm),
which increments a 32-bit time accumulator.  The MPU-401 timer value is sent
with each MIDI input message and the timer is then cleared.  Therefore,
@i(aintr.asm) must strip off the time from each MIDI input message and add
it to the time accumulator as well.  At all times, the current time is the
sum of the 32-bit time accumulator and the 8-bit timer inside the MPU-401.
Therefore, to get an accurate time reading, one must force the MPU-401 to
send the value of its counter.  Fortunately, there is an MPU-401 command for
this.  When a user calls @i(gettime@Index(gettime)) in @i(mpu.c), the
command is sent to the MPU-401, which responds by sending its counter value.
This value is added to the time accumulator by @i(aintr.asm).  The
@i(gettime) routine can then read the accumulator, which is now up to date,
shift it right 2 bits to divide by 4, and return the result, which is
absolute time in hundredths of seconds.

A new timer has been implemented using the real-time clock@Index(real-time
clock) on the IBM-PC.  The new timer code is @i(not) compatible with
an IBM-AT.
This new code should reduce some of the load on the MPU-401
when the host computer needs to poll @i(gettime).  This timer is
included in the first release of the system and can be enabled by
defining @i(TIMDIF) in @i(mpu.c) and recompiling it.  You will find
the definition of @i(TIMDIF) commented out in the file @i(mpu.c).
Notice that with a
little coding, one could get the MPU-401 to synchronize its internal clock
to an external tape machine.  In the hopes that someone will actually extend
CMT to implement tape synchronization, the code that uses the MPU-401 timer
will be left in the system even if it is disabled.

@subsection(System exclusive messages)
In an earlier implementation of CMT, incoming system exclusive
messages@Index(system exclusive messages) were put into the buffer just like
other MIDI messages.  I then discovered that an IBM XT could not empty the
buffer as fast as the messages came in.  This is fine if messages are short,
but some system exclusive messages are much longer than the buffer.  Making
the buffer bigger would penalize programs that are not interested in system
exclusive messages by taking up lots of space.  Consequently, the current
system does not put system exclusive messages in the buffer.  Instead, there
is a call (@i(midi@ux(  )buffer@Index(midi@ux(  )buffer))) through which a
user of @i(mpu.c) can provide a special buffer to be used by system
exclusive messages.  Most programs use the MPU-401 to filter out these
messages, so what to do with them is not usually an issue.  If exclusive
messages are enabled and no buffer is supplied for them, @i(mpu.c) throws
them away.

@section(Command line parsing)
The module @i(cmdline.c) implements an interface to the command line,
that is, the line that the user types to the operating system to start
a program.  There are several conventions enforced by @i(cmdline.c).
A command line must consist of three types of entities which must be
separated by one or more blanks.  In order to describe them, I have
given the entities names, although they are not terribly mnemonic:
@begin(enumerate)
A @i(switch) is a sequence of non-space characters that begins with
a dash (``-'').  Normally, a switch is used to enable or disable some
function in the program.  An example is the  ``-help'' switch, which
causes most CMT programs to print a help message and quit.

An @i(option) looks just like a switch except it is followed by another
sequence of non-space characters that do not begin with a dash.  Options
are used to give additional information to the program.  An example is
``-tune myscale.tun'', which tells @i(mpu.c) to load the tuning
definition file @i(myscale.tun).

An @i(argument) is a sequence of non-space characters that do not begin
with a dash.  Arguments are usually pieces of information the program
expects.  An example is in the command line ``adagio bach'': ``Adagio''
is the command name, and ``bach'' is an argument, which in this case
specifies a score file.
@end(enumerate)
Both switches and options are referred to in earlier chapters as
``command line options''.
Switches, options, and arguments can be placed in the command line in
any order, with the exception that arguments are numbered in the order
in which they appear from left to right.
An example of a command with several switches and arguments is:
@begin(example)
transcri -tune myscale -control off -block foo
@end(example)
Taken one at a time, the entities in this command line are:
@begin(itemize)
@i(transcri) is the name of the command.

@i(-tune myscale) is an option specifying a tuning definition file.

@i(-control off) is an option telling transcri to ignore control change
data.

@i(-block) is a switch which turns off the internal MIDI thru of the MPU-401.

@i(foo) is an argument specifying a score file.  Notice that ``-block foo''
has the same syntax as an option, but since @i(-block) is a switch, @i(foo)
must be an argument.
@end(itemize)

  This module provides the following functions:
@begin(itemize)
Syntax checking.  The entire line is checked to make sure every token
is recognized.

Testing for switches and options.  Functions are provided to test whether
the user typed a switch or option.

Argument access.  Command line arguments can be retrieved by number.
(The first argument after the command is argument number 1.)

Consistency checking.  All legal switches or options must be declared when
the @i(cmdline.c) module is initialized.  Attempts to look up a switch
or option that was not part of the initial declaration are detected and
cause an error message.
@end(itemize)

The only aspect of @i(cmdline.c) that is likely to be confusing is the
initialization and consistency checking.  Notice that 
a switch followed by an argument is syntactically identical to an option.
Thus, @i(cmdline.c) must know in advance the complete list of switches and
options in order to parse the command line.  It is illegal for a switch
to have the same name as an option, because this could lead to an ambiguous
parse.  Therefore, @i(cmdlne.c) must be initialized with a list of
switches and a separate list of options.

This is a somewhat error-prone operation because @i(cmdline.c) can be (and is)
used by many modules, but there is no convenient way in C for these modules
to make known what switches and options they look for.  For example, suppose
you write a program that uses  @i(cmdline.c) and @i(mpu.c).  You initialize
@i(cmdline.c), passing it several switches and the ``-aardvark'' option, but
you forget that @i(mpu.c) looks for a ``-tune'' option.  Consequently,
@i(cmdline.c) does not have enough information to properly parse the command
line.  Even worse, your program is booby-trapped!  As soon as some poor sole
tries to use the ``-tune'' option, @i(cmdline.c) will report a syntax error.
As far as it knows, ``-tune'' is not a valid switch or option.

Fortunately, the situation is not as bad as it looks, because @i(cmdline.c)
performs some internal consistency checks automatically.  What will happen
in this example is that as soon as @i(mpu.c) is initialized, it will go out
and look for the ``-tune'' option by calling the appropriate routine in
@i(cmdline.c) (@i[cl@ux(  )option()], for example).  Before
searching the command line, @i(cmdline.c) checks to
see if ``-tune'' is in its list of valid options.  It is not, so
@i(cmdline.c) prints an error message:
@begin(example)
internal error detected by cmdline module:
          '-tune' should be in valid lists
@end(example)
Thus you, the implementor, will get a clear message that you made a
mistake, and you will get the message whether or not ``-tune'' appears
on the command line.  If it is not obvious where the message came from,
you can use @i(grep) to search the source code of all the modules you
use for the string ``-tune''.  That would tell you that ``-tune'' is
an option used by the @i(mpu.c) module.

In the best of all worlds, it should not be necessary
for the implementor to keep track of options and switches used in other
modules, but this is the best compromise I could come up with.

@section(The record module)
Two programs in CMT record data from the keyboard and write Adagio
scores.  These are the @i(transcri) and @i(record) programs.  Both
programs use the @i(record.c) module to do most of the work.
The interface to @i(record.c) is fairly simple.  The routine @i(rec@ux(  )init)
is called to initialize the module.  Then @i(rec@ux(  )poll) is called
(frequently) for the duration of the recording ``session''.  This
routine reads data from the input buffer and stores it in a data
structure along with timing information.  When the recording is
finished, @i(rec@ux(  )final) is called to write the data to a file.

Several tricks are used to improve the performance of @i(record.c).
First, the program does not try to record a time for every input
event.  This would waste processor time when bursts of data arrive.
Instead, the current time is recorded only when something is found
in the input buffer.  Then the buffer is read until it is empty and
all buffered events are assigned the same time.  Thus, buffer reading
occurs at maximum speed when bursts arrive.

A second optimization deals with the form of the output file.
It has been found that a DX7 will send large amounts of
control information, particularly aftertouch data.  The @i(rec@ux(  )final)
routine filters the recorded data so that a given control will not
change more than once per 0.01 second time period.  This reduces the
size of the output file, reduces the amount of data to be stored
by the Adagio program, and has a minimal impact on the accuracy of
reproduction.

